/**
 * Copyright (C) 2001-2002
 * - France Telecom R&D
 * - Laboratoire Logiciels, Systemes, Reseaux - UMR 5526, CNRS-INPG-UJF
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 1.0
 *
 * Authors:
 *
 */

package org.objectweb.perseus.concurrency.optimistic;

import org.objectweb.perseus.concurrency.lib.Semaphore;
import org.objectweb.perseus.concurrency.lib.TimeStamp;
import org.objectweb.perseus.concurrency.api.ConcurrencyException;
import org.objectweb.perseus.concurrency.api.ConcurrencyManager;
import org.objectweb.perseus.dependency.api.DependencyGraph;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.util.monolog.api.Logger;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

/**
 * @author E. Bruneton
 */
public abstract class OptimisticConcurrencyManager
        implements ConcurrencyManager, BindingController {

	public final static String DEPENDENCY_GRAPH_BINDING = "dependency-graph";

	private Semaphore sem;

    protected Logger logger = null;

	private DependencyGraph dg = null;

	/**
	 * A map associating time stamps to all oids in all active contexts.
	 */
	protected Map timeStamps;

	public OptimisticConcurrencyManager() {
		super();
		sem = new Semaphore();
		timeStamps = new HashMap();
	}

	/**
	 * @param ctx is the context using the resource
	 * @param resourceId is the resource identifier
	 * @param timeStamp is the TimeStamp instance matching the resource
	 * @param hints uses to give information on access
	 * @param isDirtyBefore indicates if the object was already dirty before
	 * the access.
	 * @param isWrite indicates if the access is a write access 
	 * @return the resource state to use
	 * @throws ConcurrencyException it is not possible to create a State for
	 * this resource identifier. One reason can the unavailibity of the
	 * resource.
	 */
	protected abstract Object getState(Object ctx,
									   Object resourceId,
									   TimeStamp timeStamp,
									   Object hints,
									   boolean isDirtyBefore,
									   boolean isWrite) throws ConcurrencyException;

    //IMPLEMENTATION OF THE UserBindingControler INTERFACE //
    //------------------------------------------------------//

	public String[] listFc() {
		return new String[] {DEPENDENCY_GRAPH_BINDING};
	}

	public Object lookupFc(String s) {
		if (DEPENDENCY_GRAPH_BINDING.equals(s)) {
			return dg;
		}
		return null;
	}

	public void bindFc(String s, Object o) {
		if ("logger".equals(s)) {
			logger = (Logger) o;
		} else if (DEPENDENCY_GRAPH_BINDING.equals(s)) {
			dg = (DependencyGraph) o;
		}
	}

	public void unbindFc(String s) {
		if (DEPENDENCY_GRAPH_BINDING.equals(s)) {
			dg = null;
		}
	}

	//IMPLEMENTATION OF THE ConcurrencyManager INTERFACE //
	//---------------------------------------------------//

	public void begin(Object ctx) {
		// does nothing
	}

	public boolean validate(Object ctx) {
//TODO: system will blocked if caller of validate never calls finalize
//TODO:		(e.g. the app crashes or is killed)
		sem.P();
		Iterator tss = timeStamps.values().iterator();
		while (tss.hasNext()) {
			TimeStamp timeStamp = (TimeStamp) tss.next();
			if (!timeStamp.validate(ctx)) {
				return false;
			}
		}
		tss = timeStamps.values().iterator();
		while (tss.hasNext()) {
			TimeStamp timeStamp = (TimeStamp) tss.next();
			timeStamp.finalize(ctx);
		}
		return true;
	}

	public void finalize(Object ctx) {
		if (sem.getUser() != Thread.currentThread()) {
			throw new RuntimeException("Bad use of the concurrency manager: the validate must be call before the finalize method");
		}
		try {
			Iterator tsIt = timeStamps.values().iterator();
			while (tsIt.hasNext()) {
				TimeStamp timeStamp = null;
				timeStamp = (TimeStamp) tsIt.next();
				if (timeStamp.close(ctx)) {
						tsIt.remove();
				}
			}
//TODO: system will blocked if caller of validate never calls finalize
//TODO:		(e.g. the app crashes or is killed)
		} finally {
			sem.V();
		}
	}

	public void abort(Object ctx) {
		finalize(ctx);
	}

	/**
     * @return the TimeStamp.
     */
    public Object readIntention(Object ctx, Object resourceId, Object hints)
            throws ConcurrencyException {
		sem.P();
		TimeStamp timeStamp = getTimeStamp(resourceId);
		boolean isDirtyBefore = timeStamp.isDirty(ctx);
		timeStamp.P();
		timeStamp.readIntention(ctx);
		sem.V();
		try {
			return getState(ctx, resourceId, timeStamp, hints, isDirtyBefore, false);
		} catch(ConcurrencyException e) {
			removeUnavailableResource(ctx, resourceId, timeStamp);
			throw e;
		} finally {
			timeStamp.V();
		}
	}

    /**
     * @return the TimeStamp.
     */
	public Object writeIntention(Object ctx, Object resourceId, Object hints)
            throws ConcurrencyException {
		sem.P();
		TimeStamp timeStamp = getTimeStamp(resourceId);
		boolean isDirtyBefore = timeStamp.isDirty(ctx);
		timeStamp.P();
		timeStamp.writeIntention(ctx);
		sem.V();
		try {
			return getState(ctx, resourceId, timeStamp, hints, isDirtyBefore, true);
		} catch(ConcurrencyException e) {
			removeUnavailableResource(ctx, resourceId, timeStamp);
			throw e;
		} finally {
			timeStamp.V();
		}
	}

	private void removeUnavailableResource(Object ctx,
										  Object resourceId,
										  TimeStamp timeStamp) {
		//The instance does not exist in the data support, remove the lock
		if (timeStamp.close(ctx)) {
			timeStamp.V();
			sem.P();
			timeStamp.P();
			if (timeStamp.close(ctx)) {
				timeStamps.remove(resourceId);
			}
			sem.V();
		}
	}

	//OTHER METHODS //
	//--------------//

	/**
	 * Returns the lock associated to the given oid.
	 * @param oid an object identifier.
	 * @return the lock associated to the given oid.
	 */
	private TimeStamp getTimeStamp(Object oid) {
		TimeStamp timeStamp = (TimeStamp) timeStamps.get(oid);
		if (timeStamp == null) {
			timeStamp = new TimeStamp();
			timeStamps.put(oid, timeStamp);
		}
		timeStamp.reserve();
		return timeStamp;
	}
}
